use std::ops::Deref;

use super::*;
use crate::core::*;

pub trait StaticPairSerializable {
    type SA: Serializable;
    type SB: Serializable;

    fn s_elements(&self) -> (&Self::SA, &Self::SB);
}

pub trait StaticPair<'a, Ctx: 'a + Context>:
    'a + StaticPairSerializable<SA = Self::A, SB = Self::B>
{
    type FactoryData: 'a + Clone;
    type A: Mentionable<'a, Ctx, Fctr = Self::FA>;
    type B: Mentionable<'a, Ctx, Fctr = Self::FB>;
    type FA: Factory<'a, Ctx, Mtbl = Self::A> + InlineableFactory;
    type FB: Factory<'a, Ctx, Mtbl = Self::B>;
    type ParseError: 'a + Error;

    fn factories(factory_data: &Self::FactoryData) -> (&Self::FA, &Self::FB);
    fn from_parsed(factory_data: &Self::FactoryData, a: Self::A, b: Self::B) -> Self;
    fn from_error_a(
        factory_data: &Self::FactoryData,
        error: <Self::FA as Factory<'a, Ctx>>::ParseError,
    ) -> Self::ParseError;
    fn from_error_b(
        factory_data: &Self::FactoryData,
        error: <Self::FB as Factory<'a, Ctx>>::ParseError,
    ) -> Self::ParseError;
    fn factory_data(&self) -> Self::FactoryData;
    fn elements(&self) -> (&Self::A, &Self::B);
}

pub struct StaticPairObject<SP> {
    pair: SP,
}

impl<SP> From<SP> for StaticPairObject<SP> {
    fn from(value: SP) -> Self {
        StaticPairObject { pair: value }
    }
}

impl<SP> AsRef<SP> for StaticPairObject<SP> {
    fn as_ref(&self) -> &SP {
        &self.pair
    }
}

impl<SP> Deref for StaticPairObject<SP> {
    type Target = SP;

    fn deref(&self) -> &Self::Target {
        &self.pair
    }
}

pub struct StaticPairFactory<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> {
    factory_data: SP::FactoryData,
}

impl<SP: StaticPairSerializable> Serializable for StaticPairObject<SP> {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        let (a, b) = self.pair.s_elements();
        a.serialize(serializer);
        b.serialize(serializer);
    }
}

impl<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> Mentionable<'a, Ctx> for StaticPairObject<SP> {
    type Fctr = StaticPairFactory<'a, Ctx, SP>;

    fn factory(&self) -> Self::Fctr {
        StaticPairFactory {
            factory_data: self.pair.factory_data(),
        }
    }

    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>> {
        let (a, b) = self.pair.elements();
        let mut vec = a.points();
        vec.extend(b.points());
        vec
    }
}

impl<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> Clone for StaticPairFactory<'a, Ctx, SP> {
    fn clone(&self) -> Self {
        Self {
            factory_data: self.factory_data.clone(),
        }
    }
}

impl<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> Factory<'a, Ctx>
    for StaticPairFactory<'a, Ctx, SP>
{
    type Mtbl = StaticPairObject<SP>;

    type ParseError = SP::ParseError;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        let (fa, fb) = SP::factories(&self.factory_data);
        let a: SP::A = match fa.deserialize(deserializer, resolver.clone()) {
            Ok(a) => a,
            Err(error) => return Err(SP::from_error_a(&self.factory_data, error)),
        };
        let b: SP::B = match fb.deserialize(deserializer, resolver) {
            Ok(b) => b,
            Err(error) => return Err(SP::from_error_b(&self.factory_data, error)),
        };
        Ok(StaticPairObject {
            pair: SP::from_parsed(&self.factory_data, a, b),
        })
    }

    fn unexpected_tail(&self, tail: &[u8]) -> Self::ParseError {
        let (_, fb) = SP::factories(&self.factory_data);
        SP::from_error_b(&self.factory_data, fb.unexpected_tail(tail))
    }
}

impl<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> InlineableFactory
    for StaticPairFactory<'a, Ctx, SP>
where
    SP::FB: InlineableFactory,
{
}

impl<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> FixedSizeFactory
    for StaticPairFactory<'a, Ctx, SP>
where
    SP::FA: FixedSizeFactory,
    SP::FB: FixedSizeFactory,
{
    fn size(&self) -> usize {
        let (fa, fb) = SP::factories(&self.factory_data);
        fa.size() + fb.size()
    }
}

impl<'a, Ctx: 'a + Context, SP: StaticPair<'a, Ctx>> ConstSizeFactory
    for StaticPairFactory<'a, Ctx, SP>
where
    SP::FA: ConstSizeFactory,
    SP::FB: ConstSizeFactory,
{
    const SIZE: usize = SP::FA::SIZE + SP::FB::SIZE;
}
